#ifndef Array_h
#define Array_h

// if RANGECHECK is defined, RANGE_CHECK_HANDLER should be a function that
// handles out-of-bounds problems; NULL_POINTER_HANDLER should handle
// null pointer dereferences
#ifdef RANGECHECK
void RANGE_CHECK_HANDLER( int size, int lo, int hi, int index );
void NULL_POINTER_HANDLER( int size, int pointertype );
#endif

// Class that will do bounds-checking for you if RANGECHECK is enabled.
// if you really need speed, don't define RANGECHECK, but 40% of bugs
// in C programs are caused because of pointers, so the range-checking
// might help you in debugging.

template <class T,int lo,int hi>
class Array
{
protected:
	T data[hi-lo+1];
public:
	inline T& operator [] ( int index )
	{
#ifdef RANGECHECK
		if( index < lo || index > hi )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}

	inline T operator [] ( int index ) const
	{
#ifdef RANGECHECK
		if( index < lo || index > hi )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}
};

template <class T,int lo,int hi>
class AllocArray
{
protected:
        T* data;
public:
        AllocArray()
        {
                data = new T[hi-lo+1];
#ifdef RANGECHECK
                if( !data )
                        NULL_POINTER_HANDLER(sizeof(T),2);
#endif
        }
        ~AllocArray()
        {
                delete data;
        }
	inline T& operator [] ( int index )
	{
#ifdef RANGECHECK
		if( index < lo || index > hi )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}

	inline T operator [] ( int index ) const
	{
#ifdef RANGECHECK
		if( index < lo || index > hi )
			RANGE_CHECK_HANDLER( sizeof(T), lo, hi, index );
#endif
		return data[index-lo];
	}
};

// If RANGECHECK is defined, this pointer will call your handler
// before dereferencing a null pointer!  Aren't templates cool?

template<class T>
struct Pointer
{
	T* x;
	Pointer( T* y = 0 ): x(y) {}
	Pointer( const Pointer<T>& y ): x(y.x) {}
	T* operator -> () const
	{
#ifdef RANGECHECK
		if( !x )
			NULL_POINTER_HANDLER(sizeof(T),1);
#endif
		return x;
	}
	T& operator * () const
	{
#ifdef RANGECHECK
		if( !x )
			NULL_POINTER_HANDLER(sizeof(T),0);
#endif
		return *x;
	}
	int operator ! () const { return !x; }
	operator T* () const 	{ return x; }
};

#endif
