#pragma hdrstop
#include <ctype.h>
#include <string.h>

#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <process.h>
#include <share.h>
#include <errno.h>
#include <dos.h>
#include <stdlib.h>

#include "inout.h"
#include "finout.h"

FileSystem::_ErrorHandler FileSystem::ErrorHandler = 0;

#define BUFFER_SIZE 512

fileout::fileout(): standardout( 0 )
{
        handle = -1;
	err = 0;
        binary = 0;
        buf_len = 0;
        buf_ptr = 0;
        buf = 0;
}

fileout::fileout( const char *filename, int appendflag, int binaryflag, int wide ): standardout( wide )
{
        handle = -1;
        err = 0;
        binary = binaryflag;
	buf_ptr = 0;
	buf_len = 0;
        buf = 0;

        getbuffer();

        open( filename, appendflag, binary );
}

fileout::~fileout()
{
        close();
        if( buf )       delete buf;
}

void fileout::getbuffer()
{
	if( buf )       return;

	buf_len = BUFFER_SIZE;
	buf = 0;
	while( buf_len > 16 && buf == 0 )
	{
		buf = new char[buf_len];
		if( !buf )      buf_len /= 2;
	}

	if( !buf )      buf_len = 0;
}

int fileout::open( const char *filename, int appendflag, int binaryflag )
{
	if( exists() )
		close();

	binary = binaryflag;

/******************************************************************************/
        binary = 1;             // SR inout does auto text->binary conversion
/******************************************************************************/

	handle = ::open( filename, O_WRONLY | O_CREAT |
			( appendflag? O_APPEND : O_TRUNC ) |
			( binary? O_BINARY : O_TEXT ) |
			SH_DENYWR, S_IWRITE );

	if( handle < 0 )
	{
		handle = -1;
		err = errno;
		if( err == ENOENT )
			FileSystem::HandleError( "File '%s' not openable for output.\n", filename );
		if( err == EMFILE )
			FileSystem::HandleError( "Too many files open.  See manual.\n", NULL );
                if( err == EACCES )
                        ; // Probably a LOCK on the file

		return -1;
	}

	err = 0;

        // set the pointer to 0 so that the first write will
        // write into the beginning of the buffer
        buf_ptr = 0;

        getbuffer();

        return 0;
}

int fileout::close()
{
        if( !exists() )   return 0;

        flush();
        ::close( handle );

        handle = -1;
        err = 0;

        return 0;
}

int fileout::flush(void)
{
        if( !exists() )         return 0;

        int t;

        if( buf_ptr <= 0 || !buf )      return 0;

        t = buf_ptr;
	if( write( handle, buf, buf_ptr ) < 0 )
        {
                err = errno;
                t = -1;
        }

        buf_ptr = 0;

        return t;
}

char CC fileout::_put( char c )
{
        if( !exists() )  return 0;

	getbuffer();

        if( buf_len == 0 )
        {
                if( write( handle, &c, 1 ) == -1 )
                        return 0;
        }
        else
        {
                buf[buf_ptr] = c;
                buf_ptr++;
                if( buf_ptr >= buf_len )
                        if( flush() == -1 )
                                return 0;
        }

        return c;
}

//------------------------------------------------------------------

filein::filein()
{
        handle = -1;
	err = 0;

        binary = 0;
        buf_len = 0;
        buf_ptr = 0;
        buf = 0;
        buf_overflow = -1;
}

filein::filein( const char *filename, int binaryflag )
{
        handle = -1;
        err = 0;

        binary = binaryflag;
        buf_ptr = 0;
        buf_overflow = buf_len = 0;
        buf = 0;

        getbuffer();
	open( filename, binary );
}

filein::~filein()
{
        close();
        if( buf )       delete buf;
}

void filein::getbuffer()
{
        if( buf )       return;

        buf_len = BUFFER_SIZE;
        buf = 0;
        while( buf_len > 16 && buf == 0 )
        {
                buf = new char[buf_len];
                if( !buf )      buf_len /= 2;
        }
	buf_overflow = buf_len;
        buf_ptr = buf_len;

        if( !buf )      buf_len = 0;
}

int filein::open( const char *filename, int binaryflag )
{
        if( exists() )
                close();

        binary = binaryflag;

        handle = ::open( filename, O_RDONLY |
                        ( binary? O_BINARY : O_TEXT ) |
                        SH_DENYNONE, S_IREAD );

        if( handle < 0 )
        {
                handle = -1;
		err = errno;

		if( err == ENOENT )
			; // file not found - no error message
		if( err == EMFILE )
			FileSystem::HandleError( "Too many files open.  Try increasing FILES= in CONFIG.SYS.\n", NULL );
		if( err == EACCES )
			; // Probably a LOCK on the file

		return -1;
	}

	err = 0;

	// set the pointer to the length so that the first read will
        // read into the buffer
        getbuffer();

        buf_ptr = buf_len;
        buf_overflow = buf_len;

        return 0;
}

int filein::close()
{
	if( !exists() )   return 0;

	::close( handle );

	handle = -1;
	err = 0;

	return 0;
}

char CC filein::get()
{
        char c;

        if( !exists() )  return 0;

	if( buf_len == 0 )
	{
		int status = read( handle, &c, 1 );
		if( status == 1 )
			return c;
                err = errno;
                return 0;
        }

        if( buf_ptr >= buf_overflow )
        {
                int status = read( handle, buf, buf_len );

                if( status <= 0 )
                {
                        status = 0;
                        err = errno;
                }

                if( status < buf_len )
                {
                        for( int i = status; i < buf_len; i++ )
                                buf[i] = 0;
                }

                buf_overflow = status;
                buf_ptr = 0;
        }

        if( buf_overflow <= 0 )
        {
                err = 1;
                return 0;
        }

        c = buf[ buf_ptr ];
        buf_ptr++;

        return c;
}

char CC filein::out( char c )
{
	return c;
}

void CC filein::begininput()
{
	return;
}

void CC filein::endinput()
{
	return;
}

int CC filein::keyhit()
{
        if( !exists() )
                return 0;

        return ( err == 0 );
}

int filein::operator !()
{
        if( !exists() )  return 1;

        return ( err != 0 );
}

//------------------------------------------------------------------
#ifndef NO_RECORD_FILES

recordfile::recordfile( const char *filename, int s, byte new_code, byte new_cycle )
{
	*pathname = 0;

        lastaccess = -1;

	cycle_code = new_cycle;
	if( cycle_code < 1 )
		cycle_code = 1;
	xor_code = new_code;

	err = 0;
	size = s;
	handle = -1;
	if( filename && *filename ) open( filename );
}

recordfile::~recordfile()
{
        close();
	*pathname = 0;
}

int recordfile::open( const char *filename )
{
	if( exists() )
		close();

	handle = ::open( filename,
			 O_RDWR | O_CREAT | O_BINARY | SH_DENYRW,
			 S_IREAD | S_IWRITE );

	if( handle == -1 )
	{
		err = errno;

		if( err == ENOENT )
			FileSystem::HandleError( "File '%s' not openable.\n", filename );
		if( err == EMFILE )
			FileSystem::HandleError( "Too many files open.  Try increasing FILES= in CONFIG.SYS.\n", NULL );
		if( err == EACCES )
			; // Probably a LOCK on the file
	}

	strncpy( pathname, filename, 127 );
	pathname[127] = 0;

	return 0;
}

int recordfile::close()
{
	if( !exists() )   return 0;

        ::close( handle );
        handle = -1;
	err = 0;

        *pathname = 0;

        return 0;
}

void recordfile::encode( char *data, int count )
{
	char *p = data + count * size;

	if( xor_code == 0 && cycle_code == 1 )
                return;

	for( int i = count-1; i>=0; i-- )
	{
		for( int j = size-1; j>=0; j-- )
		{
			p--;
                        *p ^= ( (xor_code + (j%cycle_code)) & 0xff );
		}
	}
}

int CC recordfile::read( long n, void *data, int count )
{
	if( !exists() )  return 0;

        if( n >= filesize() ) lastaccess = -1;
        long l = (n==lastaccess)? 0 : ::lseek( handle, size * n, SEEK_SET );
        lastaccess = n+count;

        if( l == -1 )
	{
		err = errno;
                lastaccess = -1;
                return 0;
	}

	int status = _read( handle, data, count * size );
	if( status == -1 )
	{
                err = errno;
                lastaccess = -1;
                return 0;
        }

	if( status < count * size )
	{
		err = 1;
		lastaccess = -1;
		return 0;
	}

	encode( (char *)data, count );

	return 1;
}

int CC recordfile::write( long n, void *data, int count )
{
	if( !exists() )  return 0;

        if( n >= filesize() ) lastaccess = -1;
        long l = (n==lastaccess)? 0 : ::lseek( handle, size * n, SEEK_SET );
	lastaccess = n+count;

        if( l == -1 )
        {
                err = errno;
		lastaccess = -1;
		return 0;
        }

        encode( (char *)data, count );
        int status = _write( handle, data, count * size );
	encode( (char *)data, count );

	if( status == -1 )
	{
                lastaccess = -1;
		err = errno;
		return 0;
        }

	if( status < count * size )
	{
                lastaccess = -1;
		FileSystem::HandleError( "Error writing to file '%s', disk full?\n", pathname );

                err = 1;
		return 0;
        }

        return 1;
}

int recordfile::lockrecord( long n )
{
        err = lock( handle, size * n, size );

	return ( err == 0 );
}

int recordfile::unlockrecord( long n )
{
        err = unlock( handle, size * n, size );

        return ( err == 0 );
}

long CC recordfile::filesize()
{
        if( !exists() )  return 0;

	long l = ::filelength( handle );
        if( l == -1 )
	{
                err = errno;
                return 0;
	}

        return ( l / size );
}

int CC recordfile::erase_file()
{
        char s[128];
        strcpy( s, pathname );

	close();
	int status = unlink( s );
        open( s );
	return status != -1;
}

int CC recordfile::error()
{
	int t = err;

        err = 0;
	return t;
}

#endif

void FileSystem::HandleError( const char *msg, const char *arg )
{
	char str[256];
        int j = 0;
	for( int i = 0; msg[i]; i++ )
		if( msg[i] == '%' && msg[i+1] == 's' )
		{
			strcpy( str+i, arg );
                        j += strlen(arg);
                        i++;
		}
		else
		{
                        str[j++] = msg[i];
                        str[j] = 0;
		}
	if( ErrorHandler ) ErrorHandler( str );
}

void FileSystem::InstallErrorHandler( _ErrorHandler f )
{
	ErrorHandler = f;
}

FileSystem::_ErrorHandler FileSystem::GetErrorHandler()
{
	return ErrorHandler;
}

